/**
 * Copyright © 2014-2019 Darwino, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.darwino.domino.napi.wrap;

import java.lang.reflect.Array;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.TimeUnit;

import com.darwino.domino.napi.DominoAPI;
import com.darwino.domino.napi.DominoException;
import com.darwino.domino.napi.DominoRuntimeException;

/**
 * This class wraps the Notes MQ functionality in a {@link BlockingQueue} implementation.
 * 
 * <p>Note: many methods of {@link BlockingQueue} are not applicable to the Notes MQ
 * functionality (for example, {@link BlockingQueue#peek()}), and so these methods throw
 * {@link UnsupportedOperationException}.</p>
 * 
 * <p>Additionally, methods that access the underlying Notes message queue may throw
 * {@link DominoRuntimeException}s when encountering API issues.</p>
 * 
 * <p>Batch operations (for example {@link #addAll(Collection)}) are not currently designed
 * to be thread-safe.</p>
 * 
 * @author Jesse Gallagher
 * @since 2.5.0
 */
public class NSFMessageQueue extends NSFHandle implements BlockingQueue<String> {
	public NSFMessageQueue(NSFSession session, long handle) {
		super(session, handle);
		this.capacity = 0xFFFF;
	}
	
	private int capacity;

	@Override
	protected NSFSession getParent() {
		return getSession();
	}
	
	@Override
	protected void doFree() {
		int handle = (int)this.getHandle();
		if(handle != DominoAPI.NULLHANDLE) {
			try {
				api.MQClose(handle, 0);
			} catch (DominoException e) {
				e.printStackTrace();
			}
		}
		
		super.doFree();
	}
	
	// *******************************************************************************
	// * BlockingQueue methods
	// *******************************************************************************


	@Override
	public String poll() {
		_checkRefValidity();
		
		if(isEmpty()) {
			return null;
		} else {
			try {
				return api.MQGet((int)getHandle(), 0, 0);
			} catch (DominoException e) {
				throw DominoRuntimeException.forDominoException(e);
			}
		}
	}


	@Override
	public boolean addAll(Collection<? extends String> c) {
		for(String e : c) {
			add(e);
		}
		return !c.isEmpty();
	}

	@Override
	public void clear() {
		while(size() > 0) {
			poll();
		}
	}

	@Override public String remove() {
		if(isEmpty()) {
			throw new NoSuchElementException("Message queue is empty"); //$NON-NLS-1$
		}
		return poll();
	}

	@Override
	public boolean isEmpty() {
		System.out.println("size is " + size());
		return size() < 1;
	}
	
	@Override
	public int size() {
		_checkRefValidity();
		return api.MQGetCount((int)this.getHandle());
	}

	@Override
	public Iterator<String> iterator() {
		return Arrays.asList(toArray(new String[size()])).iterator();
	}

	@Override
	public Object[] toArray() {
		return toArray(new Object[size()]);
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T> T[] toArray(T[] t) {
		Class<T> componentType = (Class<T>)t.getClass().getComponentType();
		if(!String.class.isAssignableFrom(componentType)) {
			throw new IllegalArgumentException("Class " + componentType.getName() + " is not compatible with java.lang.String"); //$NON-NLS-1$ //$NON-NLS-2$
		}
		
		T[] result;
		int size = size();
		if(t.length < size()) {
			result = (T[])Array.newInstance(componentType, size);
		} else {
			result = t;
		}
		for(int i = 0; i < size; i++) {
			result[i] = (T)poll();
		}
		return result;
	}

	@Override
	public boolean add(String e) {
		_checkRefValidity();
		
		if(size() >= capacity) {
			throw new IllegalStateException("Message queue is at its max capacity: " + capacity); //$NON-NLS-1$
		}
		
		try {
			api.MQPut((int)getHandle(), (short) 0, e, 0);
		} catch (DominoException e1) {
			throw DominoRuntimeException.forDominoException(e1);
		}
		
		return true;
	}


	@Override
	public int drainTo(Collection<? super String> c) {
		return drainTo(c, Integer.MAX_VALUE);
	}

	@Override
	public int drainTo(Collection<? super String> c, int maxElements) {
		_checkRefValidity();
		
		int count = 0;
		try {
			String message;
			do {
				message = api.MQGet((int)getHandle(), 0, 0);
				c.add(message);
				count++;
			} while(message != null && count < maxElements);
		} catch(DominoException e) {
			if(e.getStatus() == DominoAPI.ERR_MQ_EMPTY) {
				// Then we're done
			} else {
				throw DominoRuntimeException.forDominoException(e);
			}
		}
		return count++;
	}

	@Override
	public boolean offer(String e) {
		System.out.println("Looking at size " + size() + " and capacity " + capacity);
		if(size() >= capacity) {
			return false;
		} else {
			return add(e);
		}
	}

	@Override
	public boolean offer(String e, long timeout, TimeUnit unit) throws InterruptedException {
		// TODO create an implementation for this
		throw new UnsupportedOperationException();
	}

	@Override
	public String poll(long timeout, TimeUnit unit) throws InterruptedException {
		_checkRefValidity();
		try {
			return api.MQGet((int)getHandle(), DominoAPI.MQ_WAIT_FOR_MSG, (int)unit.toMillis(timeout));
		} catch(DominoException e) {
			if(e.getStatus() == DominoAPI.ERR_MQ_TIMEOUT) {
				return null;
			} else {
				throw DominoRuntimeException.forDominoException(e);
			}
		}
	}

	@Override
	public void put(String e) throws InterruptedException {
		while(!offer(e, 50, TimeUnit.MILLISECONDS));
	}

	@Override
	public int remainingCapacity() {
		return capacity - size();
	}


	@Override
	public String take() throws InterruptedException {
		String result;
		do {
			result = poll(1000, TimeUnit.MILLISECONDS);
		} while(result == null);
		
		return result;
	}

	@Override public boolean containsAll(Collection<?> c) { throw new UnsupportedOperationException(); }
	@Override public String element() { throw new UnsupportedOperationException(); }
	@Override public String peek() { throw new UnsupportedOperationException(); }
	@Override public boolean removeAll(Collection<?> c) { throw new UnsupportedOperationException(); }
	@Override public boolean retainAll(Collection<?> c) { throw new UnsupportedOperationException(); }
	@Override public boolean contains(Object e) { throw new UnsupportedOperationException(); }
	@Override public boolean remove(Object e) { throw new UnsupportedOperationException(); }
}
