/**
 * Copyright © 2014-2019 Darwino, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.darwino.domino.napi.wrap.design;

import java.util.Collections;
import java.util.Set;

import com.darwino.domino.napi.DominoAPI;
import com.darwino.domino.napi.DominoException;
import com.darwino.domino.napi.enums.ActionFlag;
import com.darwino.domino.napi.enums.ActionType;
import com.darwino.domino.napi.struct.cd.CDACTION;
import com.darwino.domino.napi.util.DominoNativeUtils;
import com.darwino.domino.napi.wrap.NSFBase;
import com.darwino.domino.napi.wrap.NSFFormula;

/**
 * This class is a wrapper of the Notes {@link CDACTION} structure.
 * 
 * @author Jesse Gallagher
 * @since 2.5.0
 */
public class NSFAction extends NSFBase {
	private final NSFBase parent;
	private final ActionType type;
	private final int iconIndex;
	private final Set<ActionFlag> flags;
	private final String title;
	private final NSFFormula hideWhenFormula;
	private final byte[] actionData;
	
	public NSFAction(NSFBase parent, CDACTION action) {
		super(parent.getAPI());
		
		this.parent = parent;
		this.type = action.getType();
		this.iconIndex = action.getIconIndex() & 0xFFFFFFFF;
		this.flags = action.getFlags();
		this.title = action.getTitle();
		this.hideWhenFormula = addChild(action.getHideWhenFormula(parent.getAPI()));
		this.actionData = action.getActionData();
	}
	
	public ActionType getType() {
		return type;
	}
	
	public int getIconIndex() {
		return iconIndex;
	}
	
	public Set<ActionFlag> getFlags() {
		return Collections.unmodifiableSet(flags);
	}
	
	public String getTitle() {
		return title;
	}
	
	public NSFFormula getHideWhenFormula() {
		return hideWhenFormula;
	}
	
	/**
	 * Returns the script associated with this action. The language of the script
	 * depends on the the type specified in {@link #getType()}.
	 * 
	 * @return the script associated with the action, as a String, or
	 * 	<code>null</code> if the action does not have an associated script
	 */
	public String getScript() throws DominoException {
		if(this.actionData.length == 0) {
			return null;
		}

		switch(type) {
		case RUN_SCRIPT:
			// Then it'll be an LMBCS string
			// TODO look for $$ACTSRC_(index) item when the LS is not present here
			// TODO figure out why it may have an errant final character (2-byte border?)
			return DominoNativeUtils.readLMBCSString(this.actionData);
		case RUN_FORMULA:
			// Then it'll be a compiled formula
			NSFFormula formula = new NSFFormula(DominoAPI.get(), this.actionData);
			try {
				return formula.getFormulaText(false);
			} catch(DominoException e) {
				throw e;
			} finally {
				formula.free();
			}
		case RUN_JAVASCRIPT:
			// TODO look for later CDEVENT record and N CDBLOBPART records
			return null;
		case SYS_COMMAND:
		case OLDSYS_COMMAND:
			// TODO make enums for standard actions
			return null;
		case PLACEHOLDER:
		case RUN_AGENT:
		default:
			return null;
		}
	}
	
	// *******************************************************************************
	// * NSFBase methods
	// *******************************************************************************
	
	@Override
	protected void doFree() {
		// NOP
	}
	@Override
	public boolean isRefValid() {
		return true;
	}
	@Override
	protected NSFBase getParent() {
		return parent;
	}
	
}
